using System;
using System.Web;
using System.Collections.Generic;
using System.Text;
using SemWeb;
using System.IO;
using System.Collections;
using com.tr.tms.abstractionLayer;
using System.Reflection;
using log4net;

namespace com.tr.tms.abstractionLayer.engine
{

    public class RDFEngine:Engine
    {
        public const string TYPE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type";
        protected static readonly ILog log = LogManager.GetLogger(typeof(RDFEngine));
        
        public RDFEngine()
        {
            log4net.Config.XmlConfigurator.Configure();
        }

        private void CreateTheClasses(MemoryStore store, Dictionary<String, CalaisObject> classesMap)
        {
            // list the classes in the Model with the predicate rdf-syntax-ns#type
            MemoryStore classesMemoryStore = new MemoryStore();
            SelectFilter filter = new SelectFilter(); // for now all wildcards
            filter.Predicates = new Entity[] { RDFEngine.TYPE };
            store.Select(filter, classesMemoryStore);
            
            //create the instance of the classes and insert them into the map
            foreach (Statement typeStatement in classesMemoryStore)
            {
                try
                {
                    //get the object - the class name
                    string classId = typeStatement.Subject.Uri;//class ID
                    string classType = typeStatement.Object.Uri;//class type
                    // Create a new instance of the class
                    CalaisObject calaisObject = new CalaisObject(classId, classType);                    
                    //add the new class to the map
                    classesMap.Add(classId, calaisObject);
                }
                catch (Exception e)
                {
                   log.Info("The class doesn't exist,subject:" + typeStatement.Subject.Uri, e); 
                    
                } 
            }
        }

        private void AddDataToClasses(MemoryStore store, Dictionary<String, CalaisObject> classesMap)
        {
            //set all the properties
            foreach (Statement statement in store)
            {
                try
                {
                    //check whether the predicate is different than 'RDF_TYPE'
                    if (statement.Predicate.Uri != RDFEngine.TYPE)
                    {
                        //get the class from the classesMap
                        string classKey = statement.Subject.Uri;
                        CalaisObject classObject;
                        if (!classesMap.TryGetValue(classKey, out classObject))
                        {
                            log.Warn("The class doesn't exist,subject:" + statement.Subject.Uri);
                            continue;
                        }
                        //get the method name form predicate
                        string fieldName = statement.Predicate.Uri;

                        //2. get the value and prepare it's type as input to the reflection method - check whether the object is a String or a Resource, 
                        //   if it is a Resource then get it from the classesMap.
                        Resource rdfResource = statement.Object;
                        //find the type of the resource class
                        if (rdfResource is Entity)
                        {
                            CalaisObject referredCalaisObject;
                            if (classesMap.TryGetValue(rdfResource.Uri, out referredCalaisObject))
                            {
                                classObject.AddReference(fieldName, referredCalaisObject);
                                referredCalaisObject.AddBackReference(fieldName, classObject);                               
                            }
                            else
                            {
                                //(rdfResource is Uri)
                                classObject.AddExternalURI(fieldName, rdfResource.Uri);
                            }
                        }
                        else
                        {//(rdfResource is Literal)
                            if (rdfResource is SemWeb.Literal)
                            {
                                classObject.AddLiteral(fieldName, ((SemWeb.Literal)rdfResource).Value);
                            }
                            else
                            {
                                classObject.AddLiteral(fieldName, rdfResource.ToString());
                            }
                        }
                    }
                }
                catch (Exception e){
                    log.Info(e.Message, e);
                }
            }
        }

        private CalaisModel CreateCalaisModel(Dictionary<String, CalaisObject> classesMap)
        {
            CalaisModel calaisModel = new CalaisModel();
            foreach (CalaisObject calaisObject in classesMap.Values)
            {
                calaisModel.AddCalaisObject(calaisObject);
            }
            
            return calaisModel;
        }

        public CalaisModel ProcessText(string data)
        {
            if (String.IsNullOrWhiteSpace(data))
            {
                log.Error("failed to create model calais due to the string is empty");
                throw new ArgumentException("failed to create model calais due to the string is empty");
            }
            MemoryStore store = createMemoryStore(data, false);
            return Process(store);
        }

        public CalaisModel ProcessStream(Stream stream)
        {
            if (stream == null)
            {
                log.Error("failed to create model calais due to the stream is empty");
                throw new ArgumentException("failed to create model calais due to the stream is empty");
            }
            MemoryStore store = createMemoryStore(stream);
            return Process(store);
        }

        public CalaisModel ProcessFile(string url)
        {
            if (String.IsNullOrWhiteSpace(url))
            {
                log.Error("failed to create model calais due to the string is empty");
                throw new ArgumentException("failed to create model calais due to the string is empty");
            }
            MemoryStore store = createMemoryStore(url, true);
            return Process(store);
        }


        protected MemoryStore createMemoryStore(string data, bool isUrl)
        {
            MemoryStore store = new MemoryStore();
            RdfXmlReader rdfXmlReader;
            if (isUrl)
            {
                rdfXmlReader = new RdfXmlReader(data);
            }
            else
            {
                byte[] byteArray = Encoding.UTF8.GetBytes(data);
                MemoryStream stream = new MemoryStream(byteArray);
                rdfXmlReader = new RdfXmlReader(stream);
            }
            store.Import(rdfXmlReader);
            return store;
        }

        protected MemoryStore createMemoryStore(Stream stream)
        {
            MemoryStore store = new MemoryStore();
            RdfXmlReader x = new RdfXmlReader(stream);
            store.Import(x);
            return store;
        }

        private CalaisModel Process(MemoryStore memoryStore)
        {
            Dictionary<String, CalaisObject> classesMap = new Dictionary<String, CalaisObject>();

            //1. Create the classes and insert them into the classesMap.
            CreateTheClasses(memoryStore, classesMap);

            //2. fill the classes with the data
            AddDataToClasses(memoryStore, classesMap);

            //3.create the calaisModel and fill it with the data
            CalaisModel calaisModel = CreateCalaisModel(classesMap);

            return calaisModel;
        }
        /*static void Main(string[] args)
        {
            CalaisModel calaisModel = CalaisModelCreator.ReadFile("C:\\Users\\benny.vaknin\\Desktop\\rdf\\dummyRdf.xml");
            Console.WriteLine("end");
        }*/
    }
}
